﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;

namespace Analyze8EX
{
    public enum VariableType : byte
    {
        Unknown = 0x00,
        OS = 0x23,
        FlashApp = 0x24
    }

    public enum HardwareVersion : byte
    {
        Unknown = 0xFF,
        TI83Plus = 0x00,
        TI83PlusSE = 0x01,
        TI84Plus = 0x02,
        TI84PlusSE = 0x03,
        TI84PCSE = 0x05,
        TI8384PCE = 0x07
    }

    class Program
    {
        private static bool _saveDataToFiles;

        static void Main(string[] args)
        {
            if (args.Length < 1 || string.IsNullOrEmpty(args[0]))
            {
                _ShowHelp("Invalid arguments");
                return;
            }

            try
            {
                const string BASE_ADDRESS = "base=";
                var saveCalls = args.Contains("-sc");
                var baseAddress = 0x000000;
                _saveDataToFiles = !args.Contains("-ns");
                var b = args.Where(a => a.StartsWith(BASE_ADDRESS));
                if (b.Any())
                {
                    int.TryParse(b.First().Substring(BASE_ADDRESS.Length), out baseAddress);
                }

                _WriteLine(string.Format("Reading file '{0}'...", args[0]), ConsoleColor.Yellow);
                var file = new FileStream(args[0], FileMode.Open);
                var data = new byte[file.Length];
                file.Read(data, 0, data.Length);
                file.Close();
                _WriteLine("File successfully read.", ConsoleColor.Yellow);
                var baseFileName = Path.GetFileNameWithoutExtension(args[0]);
                var fileName = baseFileName;

                _WriteLine("Verifying flash header...");
                var header = ASCIIEncoding.ASCII.GetString(data, 0, 8);
                if (header != "**TIFL**")
                {
                    throw new InvalidDataException("Invalid flash header: " + header);
                }

                var varName = ASCIIEncoding.ASCII.GetString(data, 0x11, data[0x10]).Trim('\0');
                _WriteLine("* Variable name: " + varName, ConsoleColor.Green);

                var varType = data[0x31];
                var varTypeAsEnum = Enum.IsDefined(typeof(VariableType), varType) ? (VariableType)varType : VariableType.Unknown;
                _WriteLine(string.Format("* Variable type: {0} ({1})", varTypeAsEnum, varType.ToString("X02")), ConsoleColor.Green);

                var varSize = Utilities.GetLEInteger(data, 0x4A, 4);
                _WriteLine(string.Format("* Variable size: {0} bytes", varSize), ConsoleColor.Green);

                if (varTypeAsEnum == VariableType.FlashApp || varTypeAsEnum == VariableType.OS)
                {
                    _WriteLine("Interpreting variable header+data...");

                    _WriteLine("Intrpreting variable header...");
                    var appHeader = new ApplicationHeader(data, 0x4E);
                    _WriteLine(string.Format("* Variable size: {0} bytes", appHeader.ApplicationSize), ConsoleColor.Green);
                    if (_saveDataToFiles)
                    {
                        fileName = baseFileName + "_all.img";
                        _WriteLine(string.Format("Saving variable header+data to {0}...", fileName), ConsoleColor.Yellow);
                        File.WriteAllBytes(fileName, appHeader.RawData);
                    }

                    foreach (var field in appHeader.SubFields)
                    {
                        switch (field.FieldId)
                        {
                            case 0x8010:
                            case 0x8110:
                                {
                                    var id = (field.Data[0] << 8) | field.Data[1];
                                    _WriteLine(string.Format("* Field: Key ID: " + id.ToString("X04")), ConsoleColor.Green);
                                    break;
                                }
                            case 0x8020:
                            case 0x8120:
                                {
                                    var version = ASCIIEncoding.ASCII.GetString(field.Data);
                                    if (field.Data.Where(t => t < 0x41).Any())
                                    {
                                        //We have unprintable characters, treat this as raw data instead
                                        version = BitConverter.ToString(field.Data);
                                    }

                                    _WriteLine("* Field: Version: " + version, ConsoleColor.Green);
                                    break;
                                }
                            case 0x8030:
                            case 0x8130:
                                {
                                    _WriteLine("* Field: Build: " +
                                        Utilities.GetLEInteger(field.Data, 0, field.Data.Length), ConsoleColor.Green);
                                    break;
                                }
                            case 0x8040:
                            case 0x8140:
                                {
                                    _WriteLine("* Field: On-calculator Name: " +
                                        ASCIIEncoding.ASCII.GetString(field.Data), ConsoleColor.Green);
                                    break;
                                }
                            case 0x80A0:
                            case 0x81A0:
                                {
                                    var type = field.Data.FirstOrDefault();
                                    var typeAsEnum = Enum.IsDefined(typeof(HardwareVersion), type) ?
                                        (HardwareVersion)type : HardwareVersion.Unknown;
                                    _WriteLine(string.Format("* Field: Minimum Hardware Version: {0} ({1})", typeAsEnum, type),
                                        ConsoleColor.Green);
                                    break;
                                }
                            case 0x80C0:
                            case 0x81C0:
                                {
                                    _WriteLine(string.Format("* Field: Fuse Compatibility: Mask: {0}, Required: {1}",
                                        field.Data[0].ToString("X02"), field.Data[1].ToString("X02")), ConsoleColor.Green);
                                    break;
                                }
                            case 0x0320:
                                {
                                    var seconds = Utilities.GetBEInteger(field.Data, 2, field.Data.Length - 2);
                                    _WriteLine("* Field: Date Stamp: " + (new DateTime(1997, 1, 1)).AddSeconds(seconds), ConsoleColor.Green);
                                    break;
                                }
                            case 0x8070:
                            case 0x8170:
                                {
                                    _WriteLine("Interpreting variable data...");
                                    var raw = field.Data;
                                    if (_saveDataToFiles)
                                    {
                                        fileName = baseFileName + "_data.img";
                                        _WriteLine(string.Format("Saving variable data to {0}...", fileName), ConsoleColor.Yellow);
                                        File.WriteAllBytes(fileName, raw);
                                    }

                                    var eheader = ASCIIEncoding.ASCII.GetString(raw, 0, 3);
                                    if (eheader == "eZ8")
                                    {
                                        _WriteLine("* Extended (eZ8) Header:", ConsoleColor.Green);

                                        var name = ASCIIEncoding.ASCII.GetString(raw, 3, 8).Trim('\0');
                                        _WriteLine("** Name: " + name);

                                        var offset = Utilities.GetLEInteger(raw, 0x12, 3);
                                        _WriteLine("** Code+Data Offset: " + offset.ToString("X06"), ConsoleColor.Green);

                                        var size = Utilities.GetLEInteger(raw, 0x15, 4);
                                        _WriteLine("** Code+Data Size: " + size.ToString("X08"), ConsoleColor.Green);

                                        var codeData = raw.Skip(offset).ToArray();
                                        if (_saveDataToFiles)
                                        {
                                            fileName = baseFileName + "_codeData.img";
                                            _WriteLine(string.Format("Saving FlashApp code+data to {0}...", fileName), ConsoleColor.Yellow);
                                            File.WriteAllBytes(fileName, codeData);
                                        }

                                        var codeOffset = Utilities.GetLEInteger(raw, 0x1B, 3);
                                        _WriteLine("** Code Offset: " + codeOffset.ToString("X06"), ConsoleColor.Green);

                                        var copyrightOffset = Utilities.GetLEInteger(raw, 0x24, 3);
                                        _WriteLine("** Copyright Offset: " + copyrightOffset.ToString("X06"));

                                        var str = raw.Skip(copyrightOffset).ToArray();
                                        var s = ASCIIEncoding.ASCII.GetString(raw, copyrightOffset, Array.IndexOf<byte>(str, 0));
                                        s = s.Substring(0, Math.Min(s.Length, 256));
                                        _WriteLine("** Copyright (max 256 chars displayed):\n\t" + s, ConsoleColor.Green);

                                        var unknownOffset = Utilities.GetLEInteger(raw, 0x27, 3);
                                        _WriteLine("** Unknown Offset: " + unknownOffset.ToString("X06"), ConsoleColor.Green);

                                        _WriteLine("Parsing relocation table...");
                                        var relocationEntries = new Dictionary<int, int>();
                                        for (int i = 0; i < (offset - 0x2A) / 6; i++)
                                        {
                                            var key = Utilities.GetLEInteger(raw, 0x2A + (i * 6), 3);
                                            var value = Utilities.GetLEInteger(raw, 0x2A + (i * 6) + 3, 3);

                                            relocationEntries.Add(key, value);
                                        }

                                        _WriteLine("** Relocation table entries: " + relocationEntries.Count, ConsoleColor.Green);
                                        if (_saveDataToFiles)
                                        {
                                            fileName = baseFileName + "_reltbl.img";
                                            _WriteLine(string.Format("Saving relocation area to {0}...", fileName), ConsoleColor.Yellow);
                                            File.WriteAllBytes(fileName, raw.Skip(0x2A).Take(relocationEntries.Count * 6).ToArray());
                                        }

                                        _WriteLine(string.Format("Performing relocation replacements " +
                                            "with base address {0}...", baseAddress.ToString("X08")));
                                        foreach (var entry in relocationEntries)
                                        {
                                            var val = Utilities.GetLEArray(entry.Value, 3);
                                            for (int i = 0; i < val.Length; i++)
                                            {
                                                codeData[entry.Key + i] = val[i];
                                            }
                                        }

                                        if (_saveDataToFiles)
                                        {
                                            fileName = baseFileName + "_codeDataRelocated.img";
                                            _WriteLine(string.Format("Saving relocated FlashApp code+data to {0}...",
                                                fileName), ConsoleColor.Yellow);
                                            File.WriteAllBytes(fileName, codeData);
                                        }

                                        if (saveCalls)
                                        {
                                            fileName = baseFileName + "_calls.txt";
                                            _WriteLine(string.Format("Saving calls to {0}...", fileName), ConsoleColor.Yellow);
                                            for (int i = 0; i < codeData.Length - 3; i++)
                                            {
                                                if (codeData[i] == 0xCD)
                                                {
                                                    var st = codeData[i + 3].ToString("X02");
                                                    st += codeData[i + 2].ToString("X02");
                                                    st += codeData[i + 1].ToString("X02");
                                                    File.AppendAllLines(fileName, new string[] { st });
                                                }
                                            }
                                        }
                                    }
                                    break;
                                }
                            case 0x0000:
                                {
                                    _WriteLine(string.Format("* Field: Filler ({0} bytes)", field.Data.Length), ConsoleColor.Green);
                                    break;
                                }
                            default:
                                {
                                    _WriteLine(string.Format("* Unknown field '{0}', data '{1}'",
                                        field.FieldId.ToString("X04"), BitConverter.ToString(field.Data)), ConsoleColor.Green);
                                    break;
                                }
                        }
                    }
                }

                _WriteLine("Done.", ConsoleColor.Yellow);
            }
            catch (Exception ex)
            {
                _ShowHelp(ex.ToString());
            }
        }

        private static void _WriteLine()
        {
            _WriteLine(null);
        }

        private static void _WriteLine(string line)
        {
            _WriteLine(line, null);
        }

        private static void _WriteLine(string line, ConsoleColor? color)
        {
            if (color.HasValue)
            {
                Console.ForegroundColor = color.Value;
            }

            if (line != null)
            {
                Console.WriteLine(line);
            }
            else
            {
                Console.WriteLine();
            }

            Console.ResetColor();
        }

        private static void _ShowHelp()
        {
            _ShowHelp(string.Empty);
        }

        private static void _ShowHelp(string errorMessage)
        {
            var asm = Assembly.GetExecutingAssembly().GetName();
            _WriteLine(string.Format("{0} v{1}", asm.Name, asm.Version.ToString()), ConsoleColor.Yellow);
            _WriteLine();
            _WriteLine(string.Format("Usage: {0} [8E* file name] [optional arguments]", asm.Name), ConsoleColor.Yellow);
            _WriteLine("\t-ns\tDo not save sections as separate files", ConsoleColor.Yellow);
            _WriteLine("\t-sc\tSave candidate ADL mode calls to text file", ConsoleColor.Yellow);
            if (!string.IsNullOrEmpty(errorMessage))
            {
                _WriteLine();
                _WriteLine("ERROR: " + errorMessage, ConsoleColor.Red);
            }
        }
    }
}
